#include "pmetrinit.h"
#include "metrinit.h"
#include "mefelinit.h"
#include "trfelinit.h"
#include "pglobalc.h"
#include "xfile.h"
#include "globalc.h"
#include "mpi.h"
#include <stdio.h>
#include <string.h>
#include <time.h>

void pmetr_init (int argc,const char *argv[], stochdriver *stochdm, stochdrivert *stochdt)
{
  long i, merr, tmp;
  char name[1100];
  XFILE *inm,*intr,*inc;

  //  creation of main objects
  Pcp = new pprobdescc;
  
  //  creation of object of parallel solver for mechanical problem
  Psolm = new psolver (Nproc,Myrank,proc_namec,nameLengthc);

  //  creation of object of parallel solver for transport problem
  Psolt = new psolver (Nproc,Myrank,proc_namec,nameLengthc);
  

  //  docasne umisteni
  if (Bar2d==NULL)   Bar2d = new barel2d;
  if (Beam2d==NULL)   Bar2d = new barel2d;
  if (Pelq==NULL)    Pelq = new planeelemlq;
  if (Peqq==NULL)    Peqq = new planeelemqq;
  if (Asymlq==NULL)  Asymlq = new axisymlq;
  if (Lhex==NULL)  Lhex = new linhex;
  
  if (Ltet==NULL)  Ltet = new lintet;
  if (Bar3d==NULL)   Bar3d = new barel3d;
  

  //  problem description of all coupled problem
  Cp = new probdescc;
  if (process_argsc(argc, argv)){
    delete Cp;
    abort();
  }
  //  problem description of mechanical part of problem
  Mp = new probdesc;
  //  problem description of thermo-hydro part of problem
  Tp = new probdesct;
  //  topology of mechanical part of problem
  Gtm = new gtopology;
  //  topology of thermo-hydro part of problem
  Gtt = new gtopology;
  //  auxiliary united topology (mechanical + thermo-hydro)
  Gtu = new gtopology;
  //  additional topological data about mechanical part of problem
  Mt = new mechtop;
  //  additional topological data about thermo-hydro part of problem
  Tt = new transtop;
  //  additional topological data about thermo-hydro part of problem
  Ct = new couptop;
  //  materials of mechanical part of problem
  Mm = new mechmat;
  //  materials of thermo-hydro part of problem
  Tm = new transmat;
  //  materials of thermo-hydro part of problem
  Cmu = new coupmatu;
  //  materials of thermo-hydro part of problem
  Cml = new coupmatl;
  //  cross-section in mechanical part of problem
  Mc = new mechcrsec;
  //  cross-section in thermo-hydro part of problem
  Tc = new transcrsec;
  //  boundary conditions and load cases in coupled problem
  Cb = new coupbclc;
  //  boundary conditions and load cases in mechanical part of problem
  Mb = new mechbclc;
  //  boundary conditions and load cases in thermo-hydro part of problem
  Tb = new transbclc;
  //  mechanical outdriver
  Outdm = new outdriverm;
  //  transport outdriver
  Outdt = new outdrivert;
  //  coupled problem outdriver
  Outdc = new outdriverc;
  
  
  //if (Myrank==0){
  //char bd;
  //scanf ("%c",&bd);
  //}
  //MPI_Barrier (MPI_COMM_WORLD);
  

  //  names of input files
  strcpy (name,argv[1]);
  sprintf (&name[strlen (argv[1])],"%d.in",Myrank+1);
  inc = xfopen (name,"r");
  if (inc==NULL){
    par_print_err(Myrank+1, proc_namec, "input file cannot be opened.",__FILE__, __LINE__, __func__);
    metr_dealloc();
    abort();
  }

  inc->maxlnsize = 1000;
  inc->warning = 1;
  inc->ignorecase = 1;
  if ((Cp->kwdsw == nokwd) || (Cp->kwdsw == kwd_outd))
    inc->kwdmode = ignore;
  else
    inc->kwdmode = sequent_mode;
  
  //  number of subdomain
  xfscanf (inc,"%ld",&Ndom);
  Ndom--;
  
  //  problem description of all coupled problem
  Cp->read (inc);
  
  //  input file for mechanical part
  strcpy (name,Cp->minfile);
  sprintf (&name[strlen (Cp->minfile)],"%d.in",Myrank+1);
  inm = xfopen(name,"r");
  if (inm==NULL){
    par_print_err(Myrank+1, proc_namec, "input file for mechanical part cannot be opened.",__FILE__, __LINE__, __func__);
    abort ();
  }  
  //  input file for thermo-hydro part
  strcpy (name,Cp->tinfile);
  sprintf (&name[strlen (Cp->tinfile)],"%d.in",Myrank+1);
  intr = xfopen(name,"r");
  if (intr==NULL){
    par_print_err(Myrank+1, proc_namec, "input file for thermo-hydro part cannot be opened.",__FILE__, __LINE__, __func__);
    abort ();
  }
  strcpy (name,argv[1]);
  sprintf (&name[strlen (argv[1])],"%d.log",Myrank+1);
  //sprintf (&name[strlen (argv)],"%d.log",Myrank+1);
  Outc = fopen (name,"w");
  if (Outc==NULL){
    par_print_err(Myrank+1, proc_namec, "log file for coupled part cannot be opened.",__FILE__, __LINE__, __func__);
    abort ();
  }
  strcpy (name,Cp->minfile);
  sprintf (&name[strlen (Cp->minfile)],"%d.log",Myrank+1);
  Out = fopen (name,"w");
  //Out = fopen ("mefel.log","w");
  if (Out==NULL){
    par_print_err(Myrank+1, proc_namec, "log file for mechanical part cannot be opened.",__FILE__, __LINE__, __func__);
    abort ();
  }
  strcpy (name,Cp->tinfile);
  sprintf (&name[strlen (Cp->tinfile)],"%d.log",Myrank+1);
  Outt = fopen (name,"w");
  //Outt = fopen ("trfel.log","w");
  if (Outt==NULL){
    par_print_err(Myrank+1, proc_namec, "log file for thermo-hydro part cannot be opened.",__FILE__, __LINE__, __func__);
    abort ();
  }

  inm->maxlnsize = 1000;
  inm->warning = 1;
  inm->ignorecase = 1;
  if ((Mp->kwdsw == nokwd) || (Mp->kwdsw == kwd_outd))
    inm->kwdmode = ignore;
  else
    inm->kwdmode = sequent_mode;

  intr->maxlnsize = 1000;
  intr->warning = 1;
  intr->ignorecase = 1;
  if ((Tp->kwdsw == nokwd) || (Tp->kwdsw == kwd_outd))
    intr->kwdmode = ignore;
  else
    intr->kwdmode = sequent_mode;

  //  number of subdomain can be ommitted
  xfscanf (inm,"%ld", &tmp);
  //  reading data describing the problem
  Mp->read (inm);
  
  //  number of subdomain can be ommitted
  xfscanf (intr,"%ld", &tmp);
  //  basic problem data
  Tp->read (intr);
  
  //  reading data about parallel solver
  Psolm->read (inm,Gtm,Ndom,Mp->tstorsm,Mespr);
  
  //  reading data about parallel solver
  Psolt->read (intr,Gtt,Ndom,Tp->tstorkm,Mesprt);
  
  if ((Cp->tprob==growing_par_coupl_mech_trans) || (Cp->tprob==par_coupl_mech_trans)){
    //  initial time setting
    Cp->time=Cp->timecon.starttime ();
    Tp->time=Tp->timecont.starttime ();
    Mp->time=Mp->timecon.starttime ();
  }

  if (Mp->kwdsw == kwd_full)
    inm->kwdmode = sequent_mode;
  else
    inm->kwdmode = ignore;
  
  if (Tp->kwdsw == kwd_full)
    intr->kwdmode = sequent_mode;
  else
    intr->kwdmode = ignore;
  
  if (Cp->kwdsw == kwd_full)
    inc->kwdmode = sequent_mode;
  else
    inc->kwdmode = ignore;
  

  Psolm->procdomcorr ();
  Psolt->procdomcorr ();
  

  //  input of node and element data
  Mt->read (inm);
  Tt->read (intr);
  Ct->read (inc);

  // mefel mesh control TKr
  if (Mp->tprob == growing_mech_structure){
    if (Mespr==1){
      merr=Mt->mesh_check();
      if (merr > 0){
	fprintf (stdout,"MEFEL Mesh error = %ld\n\n\n",merr);
	exit(0);
      }
    }
  }

    
  // trfel mesh control
  if (Tp->tprob == growing_np_problem || Tp->tprob == growing_np_problem_nonlin){
    if (Mesprt==1){
      merr=Tt->mesh_check();
      if (merr > 0){
	fprintf (stdout,"TRFEL Mesh error = %ld\n\n\n",merr);
	exit(0);
      }
    }
  }

  // metr mesh control
  /*
  if ((Cp->tprob==growing_par_coupl_mech_trans) || (Cp->tprob==par_coupl_mech_trans)){
    if (Mesprc==1){
      merr=Ct->mesh_check();
      if (merr > 0){
	fprintf (stdout,"METR Mesh error = %ld\n\n\n",merr);
	exit(0);
      }
    }
  }
  */

  if (Cp->tprob==fully_coupled_mech_trans){
    //  unification of two topologies
    Gtu->comptop (Gtm,Gtt);
  }
  
  //  initiation of the mechanical solver
  Psolm->initiate (Gtm,argc,argv);
  //  initiation of the transport solver
  Psolt->initiate (Gtt,argc,argv);
  

  //  reading of the array ltg (local-to-global correspondence)
  Psolm->read_ltg (inm,Gtm,Out);
  Psolt->read_ltg (intr,Gtt,Outt);
  

  //  generation of code numbers for mechanical problem
  Ndofm=Psolm->ordering (Gtm,Out,proc_namec);
  if (Mesprc==1)
    fprintf (stdout,"\n number of degrees of freedom in mechanics    %ld",Ndofm);
  //  generation of code numbers for transport problem
  Ndoft=Psolt->ordering (Gtt,Outt,proc_namec);
  if (Mesprc==1)
    fprintf (stdout,"\n number of degrees of freedom in transports    %ld",Ndoft);
  if (Cp->tprob==fully_coupled_mech_trans){
    //  generation of code numbers in both parts
    Ndofc=Gtu->gencodnum ();
    //  distribution of code numbers to local topologies
    Gtu->cndistr (Gtm,Gtt);
    Ndofm=Ndofc;
    Ndoft=Ndofc;
    if (Mesprc==1)
      fprintf (stdout,"\n total number of degrees of freedom    %ld",Ndofc);
  }

  //read trfel function
  if (Tp->tprob == growing_np_problem || Tp->tprob == growing_np_problem_nonlin){
    Gtt->read_gf (intr);
  }

  //  input of material characteristics
  Mm->read (inm);
  //  definition of variables ncompstr on intergation points
  //  definition of variables ssst on intergation points
  Mm->init_ip_1 ();
  

  Tm->read (intr);

  //int. points allocation and initiation
  Tm->intpointalloc ();
  Tm->intpointinit ();

  
  if (Cp->tprob==fully_coupled_mech_trans){
    Cmu->read (inc);
    Cml->read (inc);
  }  
  
  //  input of cross section characteristics
  Mc->read (inm);
  Tc->read (intr);
  

  //  input of loading
  Mb->read (inm);
  Tb->read (intr);
  
  //read mefel function
  if (Mp->tprob == growing_mech_structure){
    Gtm->read_gf (inm);
    Mt->alloc_growstr ();
    Gtm->nm = new long [Gtm->nn];
  }

  //  prescribed values on elements investigation
  //Tt->elemprescvalue();
  //  searching for elements with nodes where source is defined
  Tb->elemsource ();

  if (Cp->tprob==fully_coupled_mech_trans){  
    //  left and right hand sides
    Lsrsc = new lhsrhsc;
    Lsrsc->alloc ();
    //Lsrsc->initcond (inc);
    
    Lsrs = new lhsrhs;
    Lsrs->lhsi=Lsrsc->lhsi;;
    Lsrs->initcond (inm);

    Lsrst = new lhsrhst;     
    Lsrst->lhsi=Lsrsc->lhsi;
    Lsrst->initcond (intr);
  }
  if ((Cp->tprob==par_coupl_mech_trans) || (Cp->tprob==growing_par_coupl_mech_trans)){
    //  object for left and right hand side of MEFEL problem
    Lsrs = new lhsrhs;
    Lsrs->alloc ();
    Lsrs->initcond (inc);
    
    //  object for left and right hand side of TRFEL problem
    Lsrst = new lhsrhst;
    Lsrst->alloc ();
    Lsrst->initcond (intr);

    //  allocation of array for temperature
    Mm->tempr = new double [Mm->tnip];
    //  allocation of array for initial temperature
    Mm->inittempr = new double [Mm->tnip];
    //  allocation of array for moisture
    Mm->moist = new double [Mm->tnip];

    for (i=0;i<Mm->tnip;i++){
      Mm->tempr[i]=0.0;
      Mm->inittempr[i]=0.0;
      Mm->moist[i]=0.0;
    }
  }
  
  Mb->alloc_sumcomp ();  // arrays for resultant force components
  if (Mm->csol)
    Gtm->comp_domain_sizes (); // maximum domain sizes used consolidation model

  // input of output description
  //metr
  if ((Cp->kwdsw == nokwd) || (Cp->kwdsw == kwd_pd))
    inc->kwdmode = ignore;
  else
    inc->kwdmode = sequent_mode;

  Outdc->read(inc);

  //mefel
  if ((Mp->kwdsw == nokwd) || (Mp->kwdsw == kwd_pd))
    inm->kwdmode = ignore;
  else
    inm->kwdmode = sequent_mode;

  Outdm->read(inm);

  //trfel
  if ((Tp->kwdsw == nokwd) || (Tp->kwdsw == kwd_pd))
    intr->kwdmode = ignore;
  else
    intr->kwdmode = sequent_mode;

  Outdt->read(intr);
  

  //  names of output files
  strcpy (name,Outdm->outfn);
  sprintf (&name[strlen (Outdm->outfn)],"%d.out",Myrank+1);
  strcpy (Outdm->outfn, name);
  if (Outdm->gf != grfmt_no)
  {
    strcpy (name,Outdm->outgrfn);
    sprintf (&name[strlen (Outdm->outgrfn)],"%d",Myrank+1);
    strcpy (Outdm->outgrfn, name);
  }
  if (Outdm->ndiag > 0)
  {
    strcpy (name,Outdm->outdiagfn);
    sprintf (&name[strlen (Outdm->outdiagfn)],"%d.dat",Myrank+1);
    strcpy (Outdm->outdiagfn, name);
  }  

  //  names of output files
  strcpy (name,Outdt->outfn);
  sprintf (&name[strlen (Outdt->outfn)],"%d.out",Myrank+1);
  strcpy (Outdt->outfn, name);
  if (Outdt->gf != grftt_no)
  {
    strcpy (name,Outdt->outgrfn);
    sprintf (&name[strlen (Outdt->outgrfn)],"%d",Myrank+1);
    strcpy (Outdt->outgrfn, name);
  }
  if (Outdt->ndiag > 0)
  {
    strcpy (name,Outdt->outdiagfn);
    sprintf (&name[strlen (Outdt->outdiagfn)],"%d.dat",Myrank+1);
    strcpy (Outdt->outdiagfn, name);
  }  


  //  zacatek TKo
  //  definition of nonlocal models and models for temperature
  Mm->init_ip_2 ();
  //  konec TKo

  //  nodes allocation
  Tt->alloc_nodes ();
  //  allocation of arrays defined on nodes
  Mt->alloc_nodes_arrays ();
  //  number of components in arrays used in outdriver
  //  maximum number of components in array strain/stress on nodes and elements
  Mt->give_maxncompstr(Mm->max_ncompstrn, Mm->max_ncompstre);
  //  maximum number of components in array other on nodes and elements
  Mt->give_maxncompo(Mm->max_nncompo, Mm->max_encompo);
  //  konec TKo


  //  prescribed displacements on elements investigation
  Mt->elemprescdisp ();
  //  reactions on elements investigation
  if (Mp->reactcomp==1){
    Mt->comreac ();
  }

  //  computation of initial values at integration points from initial nodal values
  if (Mb->ico > 0)
    Mb->inicipval();
    
  if (Mp->matmodel == nonlocal){ // NUTNO PREDELAT PRO PARALELNI ULOHU
    //  list of adjacent elements
    Gtm->adjacelem (Out);
    //  list of adjacent integration points was not restored 
    //  from file it is necessary to create it from scratch
    if (restore_adjacip() == 0)
      adjacip ();
    save_adjacip();
    ipvolume ();
  }

  //  allocation of eigenstrains
  //  eigenstrains contain strains generated by coupling terms
  Mm->alloceigstrains ();
  Mm->alloctempstrains ();
  Mm->alloceigstresses ();

  //Mt->nodedisplacements (); //tohle jeste rozmyslet??!!
  
  xfclose (inc);
  xfclose (inm);
  xfclose (intr);
  
  Pcp->shift_indices ();
}

void metr_dealloc()
{
  delete Pcp;
  delete Psolm;
  delete Psolt;

  delete Pelq;
  delete Asymlq;
  delete Bar2d;
  // delete Barq2d;
  
  delete Cp;
  delete Mp;
  delete Tp;
  delete Gtm;
  delete Gtt;
  delete Gtu;
  delete Mt;
  delete Tt;
  delete Ct;
  delete Mm;
  delete Tm;
  delete Cmu;
  delete Cml;
  delete Mc;
  delete Tc;
  delete Cb;
  delete Mb;
  delete Tb;
  delete Outdm;
  delete Outdt;
  delete Outdc;
}
